#include "llvm/Support/TypedError.h"
#include "llvm/Support/raw_ostream.h"

using namespace llvm;

class MyError : public TypedErrorInfo<MyError> {
public:
  void log(raw_ostream &OS) const override {
    OS << "MyError";
  }
};

class MySubError : public TypedErrorInfo<MySubError, MyError> {
public:
  MySubError(int ExtraContext) : ExtraContext(ExtraContext) {}
  void log(raw_ostream &OS) const override {
    OS << "MySubError - Extra context: " << ExtraContext;
  }
private:
  int ExtraContext;
};

class MyAltSubError : public TypedErrorInfo<MyAltSubError, MyError> {
public:
  void log(raw_ostream &OS) const override {
    OS << "MyAltSubError";
  }
};

TypedError foo(bool RaiseError, int SubErrorContext) {
  if (RaiseError) {
    if (SubErrorContext != 0)
      return make_typed_error<MySubError>(SubErrorContext);
    else
      return make_typed_error<MyError>();
  }

  return TypedError();
}

TypedError recoveryDemo() {

  // Test for "recovering" from errors:
  std::pair<bool, bool> Inputs[] =
    { { false, 0 }, { false, 0 }, // Two non-errors.
      { true, 7 }, { true, 11 }, // Two MySubErrors.
      { true, 0 }, { false, 0 } // One MyError, and a final non-error.
    };

  for (auto &p : Inputs) {
    if (auto Err = foo(p.first, p.second)) {
      if (Err.isA<MySubError>()) {
        outs() << "Encountered MySubError. Skipping.\n";
        consumeTypedError(std::move(Err));
        continue;
      } else
        return Err; // Bail out if we encounter an error we don't recognize.
    }
    outs() << "Processing " << p.first << ", " << p.second << "\n";
  }

  return TypedError();
}

void handle(TypedError Err) {
  catchAllTypedErrors(std::move(Err),
    // Handle 'MyAltSubError' specially:
    handleTypedError<MyAltSubError>([&](std::unique_ptr<MyAltSubError> M) {
      outs() << "Caught alt-sub-error:\n";
      M->log(outs());
      outs() << "\n";
      return TypedError();
    }),
    // Handle all other errors.
    handleTypedError<MyError>([&](std::unique_ptr<MyError> M) {
      outs() << "Caught error:\n";
      M->log(outs());
      outs() << "\n";
      return TypedError();
    })
  );
}

int main(int argc, char *argv[]) {

  // Demo the basic case:
  outs() << "Returning 'Success':\n";
  handle(foo(false, 0));
  outs() << "\nReturning 'MyError':\n";
  handle(foo(true, 0));
  outs() << "\nReturning 'MySubError':\n";
  handle(foo(true, 42));

  outs() << "\nTesting subclass catch:\n";
  handle(make_typed_error<MyAltSubError>());

  // Demo error "recovery":
  outs() << "\nTesting error recovery:\n";
  handle(recoveryDemo());

  outs() << "\n";
  return 0;
}
